with Shardware; use Shardware;

with Ada.Text_IO; use Ada.Text_IO;
with Ada.Command_Line; use Ada.Command_Line;
with Ada.Directories;

with Interfaces; use Interfaces;

procedure sillymachine is

  F : Bin_IO.File_Type;
  FileName : String (1 .. 255) := (others => ascii.nul);
  FileNameLength : Integer := 0;
  ValidHeader : ByteArr (0 .. 15);
  Arr : ByteArr (0 .. 15);
  I : Integer := 0;
  Value : Byte;

  ExecSize : Integer := 0;
  MemorySize : Integer := 1048576; -- default memory is 1 mb (TODO: figure out why 8 mb produces STORAGE_ERROR)

  PC : Integer := 0; -- Program Counter
  Instruction : ByteArr (0 .. 15);

  RegisterI : RegArrI (0 .. 27);
  RegisterF : RegArrF (0 .. 27);
  RegisterU : RegArrU (0 .. 27);

begin

  
  ValidHeader (0) := 6;
  ValidHeader (1) := 148;
  ValidHeader (2) := 32;
  ValidHeader (3) := 118;
  ValidHeader (4) := 101;
  ValidHeader (5) := 114;
  ValidHeader (6) := 121;
  ValidHeader (7) := 95;
  ValidHeader (8) := 115;
  ValidHeader (9) := 105;
  ValidHeader (10) := 108;
  ValidHeader (11) := 108;
  ValidHeader (12) := 121;
  ValidHeader (13) := 0;
  ValidHeader (14) := 0;
  ValidHeader (15) := 0;

  if Argument_Count = 0 then
    New_Line(1);
    Put_Line("Usage: sillymachine <sm binary file> <number of megabytes for wanted memory (optional)>");
    New_Line(1);
    goto THE_END;
  end if;
  
  FileNameLength := Argument (1)'Length;
  FileName(1 .. FileNameLength) := Argument (1);

  Bin_IO.Open(F, Bin_IO.In_File, FileName);
  while not Bin_IO.End_Of_File (F) loop
    Bin_IO.Read(F, Value);
    
    if ValidHeader (I) = Value then
      null;
    else
      New_Line(1);
      Put_Line("Error: Please use a valid sm binary file");
      New_Line(1);
      goto THE_END;
    end if;
    I := I + 1;
    if I = 16 then
      goto END_VALID;
    end if;

  end loop;
  << END_VALID >>
  Bin_IO.Close(F);
  Put_Line("Header passed");

  if Argument_Count > 1 then
    MemorySize := 1048576 * Integer'Value(Argument (2));
  end if;

  ExecSize := Integer(Ada.Directories.Size(FileName));
  MemorySize := MemorySize + ExecSize;

  declare
    MemoryArr : ByteArr (0 .. MemorySize);
  begin
    for i in 0 .. 27 loop -- zero initialise registers
      RegisterI (i) := 0;
      RegisterF (i) := 0.0;
      RegisterU (i) := 0;
    end loop;

    for i in 0 .. MemorySize loop -- zero initialise memory
      MemoryArr (i) := 0;
    end loop;
    
    I := 0;
    Bin_IO.Open(F, Bin_IO.In_File, FileName);
    while not Bin_IO.End_Of_File (F) loop
      Bin_IO.Read(F, Value);
      MemoryArr(I) := Value;
      I := I + 1;
    end loop;
    Bin_IO.Close(F);

    PC := 64; -- initialise program counter to entry point
    while True loop
      for i in 0 .. 15 loop -- load the instruction
        Instruction (i) := MemoryArr (PC + i);
      end loop;

      case My_ShiftU(Instruction (0), 8) + Unsigned_32(Instruction (1)) is -- execute the instruction
        when 0 => null;                                               -- NOP
        when 1 => null;                                               -- TODO: TRAP
        when 2 => MovU(RegisterU(Integer(Instruction (2))),
          Bytes_To_U(Instruction));                                     -- Move Unsigned_32 to Register
        when 3 => MovU(RegisterU(Integer(Instruction (2))),             -- Move Register to Register (Unsigned_32)
          RegisterU(Integer(Instruction (3))));
        when 4 => null;                                               -- TODO: Move Memory to Register
        when 5 => AddU(RegisterU(Integer(Instruction (2))),             -- Add Unsigned_32 to Register (Only positive values for now)
          Bytes_To_U(Instruction));
        when 6 => AddU(RegisterU(Integer(Instruction (2))),             -- Add Register to Register (Unsigned_32) 
          RegisterU(Integer(Instruction (3))));
        when 7 => MovI(RegisterI(Integer(Instruction (2))),             -- Move Int to Register
          Bytes_To_I(Instruction));
        when 8 => MovI(RegisterI(Integer(Instruction (2))),             -- Move Register to Register (Int)
          RegisterI(Integer(Instruction (3))));
        when 9 => null;                                               -- TODO: Move Memory to Register
        when 10 => AddI(RegisterI(Integer(Instruction (2))),             -- Add Int to Register 
          Bytes_To_I(Instruction));
        when 11 => AddI(RegisterI(Integer(Instruction (2))),             -- Add Register to Register (Int) 
          RegisterI(Integer(Instruction (3))));
        when 65535 => goto THE_END; -- exit opcode
        when others => null;
      end case;

      --Put_Line(Unsigned_32'Image(RegisterU(0)));
      --Put_Line(Unsigned_32'Image(RegisterU(1)));
      Put_Line(Integer'Image(RegisterI(0)));
      Put_Line(Integer'Image(RegisterI(1)));

      PC := PC + 16; -- increment program counter to next instruction
    end loop;

  end;

<< THE_END >>
exception
    when Name_Error => New_Line(1); Put("Error: File '"); Put(FileName); Put_Line("' does not exist!"); New_Line(1);
end sillymachine;
